# -*- coding: utf-8 -*-
#
#  test_freq.py
#  simplestats
# 
#  Created by Lars Yencken on 10-04-2009.
#  Copyright 2009 Lars Yencken. All rights reserved.
#

import unittest
import doctest

import freq

#----------------------------------------------------------------------------#

def suite():
    testSuite = unittest.TestSuite((
            unittest.makeSuite(FreqDistTestCase),
            unittest.makeSuite(CondFreqDistTestCase),
            doctest.DocTestSuite(freq),
        ))
    return testSuite

#----------------------------------------------------------------------------#

class FreqDistTestCase(unittest.TestCase):
    """
    This class tests the FreqDist class. 
    """
    def setUp(self):
        pass

    def testBasic(self):
        x = freq.FreqDist()
        x.inc('dog')
        self.assertEqual(x.candidates(), [('dog', 0.0)])
        self.assertEqual(x.prob('dog'), 1.0)
        x.inc('cat')
        self.assertEqual(x.prob('dog'), 0.5)
        self.assertEqual(x.prob('cat'), 0.5)
    
    def tearDown(self):
        pass

#----------------------------------------------------------------------------#

class CondFreqDistTestCase(unittest.TestCase):
    def setUp(self):
        model = freq.ConditionalFreqDist()
        model.inc('Breakfast', 'Cereal')
        model.inc('Breakfast', 'Toast')
        model.inc('Lunch', 'Sandwich')
        model.inc('Dinner', 'Spaghetti', 2)
        model.inc('Dinner', 'Stir-fry')
        self.model = model
        pass

    def test_dist(self):
        model = self.model
        self.assertEqual(
                set(model.candidates('Breakfast')),
                set([
                    ('Cereal', -0.69314718055994529),
                    ('Toast', -0.69314718055994529),
                ])
            )
        self.assertEqual(
                set(model.candidates('Dinner')),
                set([
                    ('Spaghetti', -0.40546510810816444),
                    ('Stir-fry', -1.0986122886681098),
                ])
            )
        self.assertEqual(
                set(model.candidates('Lunch')),
                set([('Sandwich', 0.0)]),
            )
        pass

    def testDerivativeDists(self):
        """
        Tests to_condition_dist() and to_sample_dist() method.
        """
        condition_dist = self.model.to_condition_dist()
        self.assertEqual(condition_dist._total, 6)
        self.assertEqual(condition_dist.prob('Dinner'), (3.0/6.0))
        self.assertEqual(condition_dist.prob('Lunch'), (1.0/6.0))
        self.assertEqual(condition_dist.prob('Breakfast'), (2.0/6.0))

        sample_dist = self.model.to_sample_dist()
        self.assertEqual(sample_dist._total, 6)
        self.assertEqual(sample_dist.prob('Spaghetti'), (2.0/6.0))
        self.assertEqual(sample_dist.prob('Stir-fry'), (1.0/6.0))
        self.assertEqual(sample_dist.prob('Sandwich'), (1.0/6.0))
        self.assertEqual(sample_dist.prob('Cereal'), (1.0/6.0))
        self.assertEqual(sample_dist.prob('Toast'), (1.0/6.0))
        pass

    def tearDown(self):
        pass

#----------------------------------------------------------------------------#


if __name__ == "__main__":
    unittest.TextTestRunner(verbosity=1).run(suite())

#----------------------------------------------------------------------------#

# vim: ts=4 sw=4 sts=4 et tw=78:




# -*- coding: utf-8 -*-
#
#  __init__.py
#  simplestats
# 
#  Created by Lars Yencken on 10-04-2009.
#  Copyright 2009 Lars Yencken. All rights reserved.
#

"""
A simple statistics package.
"""

from basic import mean, stddev, basic_stats
from freq import FreqDist, ConditionalFreqDist

# vim: ts=4 sw=4 sts=4 et tw=78:




# -*- coding: utf-8 -*-
#
#  comb.py
#  simplestats
# 
#  Created by Lars Yencken on 10-04-2009.
#  Copyright 2009 Lars Yencken. All rights reserved.
#

"""
Combinations and permuatations of items.
"""

from itertools import ifilter

def combinations(*combination_list):
    """
    Generates a list of all possible combinations of one element from the
    first item in combination_list, one from the second, etc. For example::

        >>> combinations([1, 2], ['dog'], ['a', 'b'])
        [(1, 'dog', 'a'), (2, 'dog', 'a'), (1, 'dog', 'b'), (2, 'dog', 'b')]
    """
    combination_list = list(combination_list[:])
    combination_list.reverse()

    first_list = combination_list.pop()
    combos = map(lambda x: (x,), first_list)

    while combination_list:
        next_level_combos = []
        for item_to_add in combination_list.pop():
            # add this item to the end of every existing combo 
            for existing_combo in combos:
                next_level_combos.append(existing_combo + (item_to_add,))

        combos = next_level_combos

    return combos

#----------------------------------------------------------------------------#

def iunique_pairs(input_list):
    return UniquePairsIterator(input_list)

class UniquePairsIterator(object):
    """
    An interator over pairings which also has a length method.

    >>> x = UniquePairsIterator([1, 2, 3])
    >>> len(x)
    3
    >>> list(x)
    [(1, 2), (1, 3), (2, 3)]
    """
    def __init__(self, input_list):
        self.i = 0
        self.j = 1
        self.input_list = sorted(input_list)
        self.list_len = len(input_list)

        if self.list_len < 2:
            raise ValueError, "input must be of length at least 2"

    def next(self):
        if self.i == self.list_len - 1 and self.j >= self.list_len:
            raise StopIteration

        item = self.input_list[self.i], self.input_list[self.j]
        self.j += 1
        if self.j >= self.list_len:
            self.i += 1
            self.j = self.i + 1

        return item

    def __len__(self):
        return self.list_len * (self.list_len - 1) / 2

    def __iter__(self):
        return self

    def __repr__(self):
        return '<UniquePairsIterator: %d items>' % len(self)

#----------------------------------------------------------------------------#

def unique_tuples(input_list, n=2):
    "Similar to combinations, but selects from the same list."
    def filter_fn(x):
        for i in xrange(n-1):
            if x[i] >= x[i+1]:
                return False
        else:
            return True

    return filter(filter_fn, combinations(*(n*[input_list])))

#----------------------------------------------------------------------------#

def iunique_tuples(input_list, n=2):
    "An iterator version of unique_tuples."
    def filter_fn(x):
        for i in xrange(n-1):
            if x[i] >= x[i+1]:
                return False
        else:
            return True

    return ifilter(filter_fn, icombinations(*(n*[input_list])))

#----------------------------------------------------------------------------#

def icombinations(*combination_lists):
    """
    As for combinations(), but returns an iterator.
    """
    combination_lists = map(list, combination_lists)
    lengths = map(len, combination_lists)
    combined = zip(combination_lists, lengths)
    n_combs = reduce(lambda x, y: x*y, lengths)

    for i in xrange(n_combs):
        item = ()
        for item_list, list_length in combined:
            i, offset = divmod(i, list_length)
            item += (item_list[offset],)
        yield item

    return

#----------------------------------------------------------------------------#

def combination_seqs(*combination_list):
    """
    As with combinations() above, except that each potential item is
    assumed to already be in sequence form. For example::

        >>> combination_seqs([(1, 2), (3, 4)], [('dog',), ('cat',)])
        [(1, 2, 'dog'), (3, 4, 'dog'), (1, 2, 'cat'), (3, 4, 'cat')]
    """
    return list(icombination_seqs(*combination_list))

#----------------------------------------------------------------------------#

def icombination_seqs(*combination_lists):
    """
    As for combinations(), but returns an iterator.

        >>> list(icombination_seqs([(1, 2), (3, 4)], [('dog',), ('cat',)]))
        [(1, 2, 'dog'), (3, 4, 'dog'), (1, 2, 'cat'), (3, 4, 'cat')]
    """
    for seq_combs in icombinations(*combination_lists):
        result = []
        for seq in seq_combs:
            result.extend(seq)
        yield tuple(result)
    return

#----------------------------------------------------------------------------#

def segment_combinations(g_string):    
    """
    Determines the possible segment combinations based on the grapheme
    string alone, in particular due to kanji placement. For example::

        >>> segment_combinations('ab')
        [('a', 'b'), ('ab',)]
    
    """
    # start out with just the first character
    segmentations = [[g_string[0]]]

    # add remaining characters one by one
    for char in g_string[1:]: 
        next_segmentation_round = []
        for segment in segmentations:
            # the new char in its own segment
            next_segmentation_round.append(segment + [char])

            # the new char as part of the previous segment
            segment[-1] += char
            next_segmentation_round.append(segment)

        segmentations = next_segmentation_round
    
    segmentations = map(tuple, segmentations)

    return segmentations

#----------------------------------------------------------------------------#

def isegment_combinations(g_string):
    """
    As for segment_combinations(), but returns an iterator.

        >>> list(sorted(isegment_combinations('ab')))
        [('a', 'b'), ('ab',)]

    Note that the order may be different.
    """
    if not g_string:
        return

    g_string_size = len(g_string)
    n_combs = 2**(g_string_size-1)

    for i in xrange(n_combs):
        current_comb = [g_string[0]]
        for j in xrange(1,g_string_size):
            i, has_boundary = divmod(i, 2)
            if has_boundary:
                current_comb.append(g_string[j])
            else:
                current_comb[-1] += g_string[j]
        yield tuple(current_comb)

    return

#----------------------------------------------------------------------------#

def inclusion_combinations(sequence):
    """
    Returns a list of all combinations of inclusion/exclusion of the
    elements of the given sequence.

    >>> inclusion_combinations([])
    [[]]
    >>> inclusion_combinations([1, 2])
    [[], [1], [2], [1, 2]]
    """
    current_combs = [[]]
    for element in sequence:
        next_set = []
        for comb in current_combs:
            next_set.append(comb + [element])

        current_combs += next_set

    return current_combs

# vim: ts=4 sw=4 sts=4 et tw=78:



# -*- coding: utf-8 -*-
#
#  stats.py
#  simplestats
# 
#  Created by Lars Yencken on 10-04-2009.
#  Copyright 2009 Lars Yencken. All rights reserved.
#

"""
This module is responsible for any general combinatoric methods, in
particular determining possible combinations of input.
"""

from math import sqrt

from errors import InsufficientData

#----------------------------------------------------------------------------#

def mean(values):
    """
    Returns the mean of a sequence of values. If the sequence is empty, raises
    an InsufficientData error.

    >>> mean([1, 2, 3])
    2.0
    """
    values_iter = iter(values)
    n = 1

    # Need at least one value.
    try:
        total = values_iter.next()
    except StopIteration:
        raise InsufficientData

    for value in values_iter:
        total += value
        n += 1

    return total / float(n)

#----------------------------------------------------------------------------#

def stddev(values):
    """
    Returns the standard deviation of a sequence of values. If less than three
    values are provided, raises an InsufficientData error.

    >>> stddev([1, 2, 3])
    1.0
    """
    values_iter = iter(values)
    try:
        value = values_iter.next()
    except StopIteration:
        raise InsufficientData

    total = value
    totalSquared = value * value
    n = 1

    for value in values_iter:
        total += value
        totalSquared += value * value
        n += 1

    # Need at least two values.
    if n < 2:
        raise InsufficientData

    n = float(n)
    return sqrt((totalSquared - total * total / n) / (n - 1))

#----------------------------------------------------------------------------#

def basic_stats(values):
    """
    Returns the mean and standard deviation of the sample as a tuple.

    >>> basic_stats([1, 2, 3])
    (2.0, 1.0)
    """
    total = 0.0
    total_sq = 0.0
    i = -1
    for i, v in enumerate(values):
        total += v
        total_sq += v * v

    n = i + 1

    if n == 0:
        raise InsufficientData

    n = float(n)
    if n > 2:
        stddev_val = sqrt((total_sq - total * total / n) / (n - 1))
    else:
        stddev_val = None

    mean_val = total / n

    return (mean_val, stddev_val)

#----------------------------------------------------------------------------#

def is_nan(x):
    """
    Returns True if the number is NaN, False otherwise.

    >>> x = 1e300
    >>> is_nan(x)
    False
    >>> inf = x*x
    >>> is_nan(inf)
    False
    >>> nan = inf - inf
    >>> is_nan(nan)
    True
    """
    x = float(x)
    if max(x, 1e10) is x and min(x, -1e10) is x:
        return True
    else:
        return False

#----------------------------------------------------------------------------#




# -*- coding: utf-8 -*-
#
#  test_agreement.py
#  simplestats
# 
#  Created by Lars Yencken on 10-04-2009.
#  Copyright 2009 Lars Yencken. All rights reserved.
#

import unittest
import doctest
import agreement

#----------------------------------------------------------------------------#

def suite():
    testSuite = unittest.TestSuite((
            unittest.makeSuite(KappaTest),
            doctest.DocTestSuite(agreement)
        ))
    return testSuite

#----------------------------------------------------------------------------#

class KappaTest(unittest.TestCase):
    def setUp(self):
        self.dataA = [1, 2, 3, 4]
        self.dataB = [2, 3, 4, 5]

    def testHighKappa(self):
        """
        Tests a high kappa value
        """
        kappaVal = agreement.kappa(self.dataA, self.dataA)
        self.assertAlmostEqual(kappaVal, 1.0)
        return
    
    def testNoAgreement(self):
        """
        Tests the no agreement case.
        """
        kappaVal = agreement.kappa(self.dataA, self.dataB)
        self.assertAlmostEqual(kappaVal, -0.23076923076923078)
        return

    def testZeroKappa(self):
        kappaVal = agreement.kappa(self.dataA, [0,0,6,6])
        self.assertAlmostEqual(kappaVal, 0)
        return

    def tearDown(self):
        pass
#----------------------------------------------------------------------------#

if __name__ == "__main__":
    unittest.TextTestRunner(verbosity=1).run(suite())

#----------------------------------------------------------------------------#

# vim: ts=4 sw=4 sts=4 et tw=78:




# -*- coding: utf-8 -*-
#
#  aggregate.py
#  simplestats
# 
#  Created by Lars Yencken on 10-04-2009.
#  Copyright 2009 Lars Yencken. All rights reserved.
#

"""
Aggregating data into bins or other approximations.
"""

_eps = 1e-8

def bins_by_data(data, n):
    """
    Puts the data into n sorted bins. Where n does not divide the length
    of the data directly, distributes the remainder as evenly as possible.
    Returns an iterator over the bins.

    @param data: A sequence of data. 
    """
    data.sort()

    assert n <= len(data), "Can't split a group more ways than its length"

    items_per_group, remainder = divmod(len(data), n)

    start_at = 0
    for i in xrange(n):
        end_at = start_at + items_per_group

        if remainder > 0:
            end_at += 1
            remainder -= 1

        yield (start_at, end_at), data[start_at:end_at]

        start_at = end_at

    return

def bins_by_increment(data, inc, key=lambda x: x[0]):
    """
    Calculates bins by range increment. Assumes data is a sequence of
    tuples, where the first tuple is the one whose range is divided up.
    """
    data = list(data)
    data.sort()

    # add _eps to the end of the range to ensure we capture that object
    start_range = key(data[0])
    end_range = key(data[-1]) + _eps

    for bin_start in frange(start_range, end_range, inc):
        bin_end = bin_start + inc

        bin_data = [x for x in data if key(x) >= bin_start and \
                key(x) < bin_end]

        yield (bin_start, bin_end), bin_data

    return

def bins_by_range(data, n, key=lambda x: x[0]):
    """
    Calculates bins by range. Assumes data is a sequence of tuples, where
    the first tuple is the one whose range is divided up.
    """
    data = list(data)
    data.sort()

    start_range = key(data[0])
    end_range = key(data[-1])
    bin_size = (end_range - start_range)/float(n)

    for i in xrange(n):
        bin_start = start_range + i*bin_size
        bin_end = start_range + (i+1)*bin_size

        # add eps to the size of the last bin to ensure we capture that
        # object
        if i == (n-1):
            use_bin_end = bin_end + _eps
        else:
            use_bin_end = bin_end

        bin_data = [x for x in data if key(x) >= bin_start and \
                key(x) < use_bin_end]

        yield (bin_start, bin_end), bin_data

    return

def frange(start, end=None, inc=None):
    """
    A range function, that does accept float increments...
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/66472

        >>> frange(1.0, 3.0, 0.5)
        [1.0, 1.5, 2.0, 2.5]
    """ 

    if end == None:
        end = start + 0.0
        start = 0.0

    if inc == None:
        inc = 1.0

    L = []
    while 1:
        next = start + len(L) * inc
        if inc > 0 and next >= end:
            break
        elif inc < 0 and next <= end:
            break
        L.append(next)
        
    return L

# vim: ts=4 sw=4 sts=4 et tw=78:



# -*- coding: utf-8 -*-
#
#  test_basic.py
#  simplestats
# 
#  Created by Lars Yencken on 10-04-2009.
#  Copyright 2009 Lars Yencken. All rights reserved.
#

import sys, unittest
import doctest

import basic

#----------------------------------------------------------------------------#

def suite():
    testSuite = unittest.TestSuite((
            unittest.makeSuite(BasicStatsTest),
            doctest.DocTestSuite(basic),
        ))
    return testSuite

#----------------------------------------------------------------------------#

class BasicStatsTest(unittest.TestCase):
    def setUp(self):
        self.dataA = [1,2,3]
        self.meanA = 2.0
        self.stddevA = 1.0

        self.dataB = [-50.4, 30.2, 0.1, 4.327]
        self.meanB = -3.94325
        self.stddevB = 33.70824
        return

    def testMean(self):
        """
        Check that mean works as expected.
        """
        self.assertAlmostEqual(basic.mean(self.dataA), self.meanA, 5)
        self.assertAlmostEqual(basic.mean(self.dataB), self.meanB, 5)
        return

    def testBadStddev(self):
        """
        Check stddev on a list of length 0.
        """
        try:
            basic.stddev([])
        except:
            pass
        else:
            assert False, "Stddev didn't raise an exception on an empty list"

        return

    def testStddev(self):
        """
        Check that stddev works as expected.
        """
        self.assertAlmostEqual(basic.stddev(self.dataA), self.stddevA, 5)
        self.assertAlmostEqual(basic.stddev(self.dataB), self.stddevB, 5)
        return

    def testOnTuples(self):
        """
        Checks that methods also work on tuples.
        """
        self.assertAlmostEqual(basic.mean(tuple(self.dataA)), self.meanA, 5)
        self.assertAlmostEqual(basic.mean(tuple(self.dataB)), self.meanB, 5)
        self.assertAlmostEqual(basic.stddev(tuple(self.dataA)), self.stddevA, 5)
        self.assertAlmostEqual(basic.stddev(tuple(self.dataB)), self.stddevB, 5)
        return

    def testBasicStats(self):
        """
        Test that the basic stats method works.
        """
        meanA, stddevA = basic.basic_stats(self.dataA)
        self.assertAlmostEqual(meanA, basic.mean(self.dataA))
        self.assertAlmostEqual(stddevA, basic.stddev(self.dataA))
        return

    def tearDown(self):
        pass

#----------------------------------------------------------------------------#


#----------------------------------------------------------------------------#

class CombinationTest(unittest.TestCase):
    def setUp(self):
        self.dataA = [1,2,3]

    def testCombinations(self):
        self.assertEqual(combinations([self.dataA, self.dataA]),
                [(1,1),(2,1),(3,1),(1,2),(2,2),(3,2),(1,3),(2,3),(3,3)])
        return

    def testICombinations(self):
        self.assertEqual(list(icombinations([self.dataA, self.dataA])),
                [(1,1),(2,1),(3,1),(1,2),(2,2),(3,2),(1,3),(2,3),(3,3)])

        return

    def testUniqueTuples(self):
        self.assertEqual(unique_tuples([1,2]), [(1,2)])
        self.assertEqual(unique_tuples([1,2,3]), [(1,2), (1,3), (2,3)])

        return

    def testInclusionCombination(self):
        self.assertEqual(
                set(map(tuple, inclusion_combinations([1,2]))),
                set(map(tuple, [[], [1], [2], [1,2]])),
            )
        self.assertEqual(inclusion_combinations([]), [[]])
        return

    def testIUniqueTuples(self):
        self.assertEqual(list(iunique_tuples([1,2])), [(1,2)])
        self.assertEqual(list(iunique_tuples([1,2,3])), [(1,2), (1,3), (2,3)])

        return

#----------------------------------------------------------------------------#

if __name__ == "__main__":
    unittest.TextTestRunner(verbosity=1).run(suite())




# -*- coding: utf-8 -*-
#
#  test_comb.py
#  simplestats
# 
#  Created by Lars Yencken on 10-04-2009.
#  Copyright 2009 Lars Yencken. All rights reserved.
#

import sys, unittest
import doctest

import comb

def suite():
    testSuite = unittest.TestSuite((
            unittest.makeSuite(CombinationTest),
            doctest.DocTestSuite(comb),
        ))
    return testSuite

#----------------------------------------------------------------------------#

class CombinationTest(unittest.TestCase):
    def setUp(self):
        self.dataA = [1,2,3]

    def testCombinations(self):
        self.assertEqual(comb.combinations(self.dataA, self.dataA),
                [(1,1),(2,1),(3,1),(1,2),(2,2),(3,2),(1,3),(2,3),(3,3)])
        return

    def testICombinations(self):
        self.assertEqual(list(comb.icombinations(self.dataA, self.dataA)),
                [(1,1),(2,1),(3,1),(1,2),(2,2),(3,2),(1,3),(2,3),(3,3)])

        return

    def testUniqueTuples(self):
        self.assertEqual(comb.unique_tuples([1,2]), [(1,2)])
        self.assertEqual(comb.unique_tuples([1,2,3]), [(1,2), (1,3), (2,3)])

        return

    def testInclusionCombination(self):
        self.assertEqual(
                set(map(tuple, comb.inclusion_combinations([1,2]))),
                set(map(tuple, [[], [1], [2], [1,2]])),
            )
        self.assertEqual(comb.inclusion_combinations([]), [[]])
        return

    def testIUniqueTuples(self):
        self.assertEqual(list(comb.iunique_tuples([1,2])), [(1,2)])
        self.assertEqual(list(comb.iunique_tuples([1,2,3])), [(1,2), (1,3),
                (2,3)])

        return

    def testSegmentCombinations(self):
        result = set([('d', 'o', 'g'), ('do', 'g'), ('d', 'og'), ('dog',)])
        self.assertEqual(
                set(comb.segment_combinations(['d', 'o', 'g'])),
                result
            )
        self.assertEqual(
                set(comb.isegment_combinations(['d', 'o', 'g'])),
                result
            )

#----------------------------------------------------------------------------#

if __name__ == "__main__":
    unittest.TextTestRunner(verbosity=1).run(suite())




# -*- coding: utf-8 -*-
#
#  agreement.py
#  simplestats
# 
#  Created by Lars Yencken on 10-04-2009.
#  Copyright 2009 Lars Yencken. All rights reserved.
#

"""
Measuring agreement between raters.
"""

from itertools import izip

from freq import FreqDist
from errors import InsufficientData

def kappa(responses_a, responses_b, potential_responses=None):
    """
    Assuming matched list of response values for each rater, determine
    their kappa value using Cohen's method.
    """
    if not responses_a or not responses_b:
        raise InsufficientData, "Need at least one response to calculate kappa"

    if len(responses_a) != len(responses_b):
        raise ValueError, "Response vectors are different lengths"
    
    # Build rater bias distributions.
    bias_a = FreqDist()
    for sample in responses_a:
        bias_a.inc(sample)

    bias_b = FreqDist()
    for sample in responses_b:
        bias_b.inc(sample)

    if potential_responses is None:
        # Detect the sample range.
        potential_responses = set(bias_a.keys()).union(bias_b.keys())
    
    # calculate p_agreement: the actual frequency of agreement
    n_agreements = 0
    n_questions = 0
    for responseA, responseB in izip(responses_a, responses_b):
        if responseA == responseB:
            # they agreed 
            n_agreements += 1

        n_questions += 1

    assert n_questions > 0
    p_agreement = n_agreements / float(n_questions)

    assert 0 <= p_agreement <= 1, "P(Agreement) should be a defined probability"

    # calculate p_expected: the agreement expected by chance
    p_expected = 0.0
    for response in potential_responses:
        p_expected += bias_a.prob(response) * bias_b.prob(response)
    
    assert 0 <= p_expected <= 1, \
        "P(Expected) should be bewteeen 0 and 1, not %.2f" % p_expected

    # calculate kappa
    kappa = (p_agreement - p_expected)/(1 - p_expected)

    return kappa

# vim: ts=4 sw=4 sts=4 et tw=78:



# -*- coding: utf-8 -*-
#
#  test_aggregate.py
#  simplestats
# 
#  Created by Lars Yencken on 10-04-2009.
#  Copyright 2009 Lars Yencken. All rights reserved.
#

import sys, unittest
import doctest

import aggregate

#----------------------------------------------------------------------------#

def suite():
    testSuite = unittest.TestSuite((
            unittest.makeSuite(AggregateTestCase),
        ))
    return testSuite

#----------------------------------------------------------------------------#

class AggregateTestCase(unittest.TestCase):
    def setUp(self):
        return

    def testBinsByData(self):
        """
        Tests splitting data into bins.
        """
        data = [3,2,5,1,7,3,0]

        expectedLabels = [(0,3), (3,5), (5,7)]
        expectedBins = [[0,1,2], [3,3], [5,7]]

        self.assertEqual(
                list(aggregate.bins_by_data(data, 3)),
                zip(expectedLabels, expectedBins)
            )

        return

    def testBinsByRange(self):
        """
        Tests splitting data into bins.
        """
        data = [3,2,5,1,9,3,0]

        expectedLabels = [(0.0,3.0), (3.0,6.0), (6.0,9.0)]
        expectedBins = [[0,1,2], [3,3,5], [9]]

        self.assertEqual(
                list(aggregate.bins_by_range(data, 3, lambda x: x)),
                zip(expectedLabels, expectedBins)
            )

        return

    def testBinsByInc(self):
        """
        Tests splitting data into bins.
        """
        data = [3,2,5,1,8,3,0]

        expectedLabels = [(0.0,3.0), (3.0,6.0), (6.0,9.0)]
        expectedBins = [[0,1,2], [3,3,5], [8]]

        self.assertEqual(
                list(aggregate.bins_by_increment(data, 3.0, lambda x: x)),
                zip(expectedLabels, expectedBins)
            )

        data = [3,2,5,1,9,3,0]

        expectedLabels = [(0.0,3.0), (3.0,6.0), (6.0,9.0), (9.0, 12.0)]
        expectedBins = [[0,1,2], [3,3,5], [], [9]]

        self.assertEqual(
                list(aggregate.bins_by_increment(data, 3.0, lambda x: x)),
                zip(expectedLabels, expectedBins)
            )

        return


    def tearDown(self):
        pass

#----------------------------------------------------------------------------#

if __name__ == "__main__":
    unittest.TextTestRunner(verbosity=1).run(suite())




# -*- coding: utf-8 -*-
#
#  setup.py
#  simplestats
# 
#  Created by Lars Yencken on 10-04-2009.
#  Copyright 2009 Lars Yencken. All rights reserved.
#

from setuptools import setup

VERSION = '0.1.1'

f = open('src/__version__.py', 'w')
f.write('# Autogenerated by setup.py\n')
f.write('version = "%s"\n' % VERSION)
f.close()

setup(
        name='simplestats',
        description="Simple statistics modules for data analysis.",
        long_description = """
        Provides basic statistics calcuations, frequency distributions, data
        approximation methods, and others.
        """,
        url="http://bitbucket.org/lars512/simplestats/",
        version=VERSION,
        author="Lars Yencken",
        author_email="lljy@csse.unimelb.edu.au",
        license="BSD",

        package_dir={'simplestats': 'src'},
        packages=['simplestats'],
    )



# -*- coding: utf-8 -*-
#
#  freq.py
#  simplestats
# 
#  Created by Lars Yencken on 10-04-2009.
#  Copyright 2009 Lars Yencken. All rights reserved.
#

"A simple frequency distribution, modelled after that in NLTK."

#----------------------------------------------------------------------------#

import sys
import bz2
import gzip
import codecs

from math import log

#----------------------------------------------------------------------------#

class FreqDist(dict):
    """
    A simple frequency distribution, and some methods to access MLE
    probability estimations based on this distribution.

        >>> x = FreqDist()
        >>> x.inc('a', 3)
        >>> x.inc('b')
        >>> x.prob('a')
        0.75
        >>> x.prob('b')
        0.25
    """
    #------------------------------------------------------------------------#
    # PUBLIC METHODS
    #------------------------------------------------------------------------#

    def __init__(self, pairSeq=None):
        """
        Constructor. Can optionally be given a sequence of (sample, count)
        pairs to load counts from.
        """
        self._total = 0

        if pairSeq is not None:
            for sample, count in pairSeq:
                self[sample] = count
                self._total += count
        return

    #------------------------------------------------------------------------#

    def total():
        doc = "The total count."
        def fget(self):
            return self._total
        return locals()
    total = property(**total())

    #------------------------------------------------------------------------#

    def inc(self, sample, n=1):
        """Add one to the count for this sample."""
        self.__setitem__(sample, self.get(sample, 0) + n)
        self._total += n
        return

    #------------------------------------------------------------------------#

    def decrement(self, sample, n=1):
        """Remove one to the count for this sample."""
        count = self[sample]
        
        if count < n:
            raise ValueError, "Can't reduce a count below zero"
        elif count == n:
            # If we reduce a count to zero, delete it.
            self._total -= n
            del self[sample]
        else:
            # Reduce it by the given amount only.
            self._total -= n
            self[sample] -= n

        return

    #------------------------------------------------------------------------#

    def remove_sample(self, sample):
        """
        Removes the sample and its count from the distribution. Returns
        the count of the sample.
        """
        count = self[sample]
        del self[sample]
        self._total -= count

        return count

    #------------------------------------------------------------------------#

    def count(self, sample):
        """Return the frequency count of the sample."""
        return self.get(sample, 0)

    #------------------------------------------------------------------------#

    def prob(self, sample):
        """Returns the MLE probability of this sample."""
        return self.get(sample, 0) / float(self._total)

    #------------------------------------------------------------------------#

    def log_prob(self, sample):
        """Returns the log MLE probability of this sample."""
        return log(self.get(sample, 0) / float(self._total))

    #------------------------------------------------------------------------#

    def candidates(self):
        """
        Returns a list of (sample, log_prob) pairs, using the log MLE
        probability of each sample.
        """
        return [
                (k, log(v/float(self._total))) \
                for (k, v) \
                in self.iteritems()
            ]

    #------------------------------------------------------------------------#

    def dump(self, filename):
        """
        Dump the current counts to the given filename. Note that symbols
        are coerced to strings, so arbitrary objects may not be
        reconstructed identically.
        """
        o_stream = sopen(filename, 'w')
        for key, count in sorted(self.iteritems(), key=lambda x: x[1],  
                reverse=True):
            key = _escape_spaces(unicode(key))
            print >> o_stream, "%s %d" % (key, count)
        o_stream.close()
        return

    #------------------------------------------------------------------------#

    def load(self, filename):
        """
        Loads counts from the given filename. Can be done for more than
        one file.
        """
        i_stream = sopen(filename, 'r')
        for line in i_stream:
            key, count = line.rstrip().split(_symbol_sep)
            key = _unescape_spaces(key)
            count = int(count)
            self.inc(key, count)
        i_stream.close()

        return

    #------------------------------------------------------------------------#

    @staticmethod
    def from_file(filename):
        """
        An alternative constructor which builds the distribution from a file.
        """
        dist = FreqDist()
        dist.load(filename)
        return dist

    #------------------------------------------------------------------------#

    def merge(self, rhs_dist):
        for sample, count in rhs_dist.iteritems():
            self.inc(sample, count)
        return

    #------------------------------------------------------------------------#
    # PRIVATE METHODS
    #------------------------------------------------------------------------#

    #------------------------------------------------------------------------#

#----------------------------------------------------------------------------#

class DefaultFreqDist(FreqDist):
    """
    A wrapper for a frequency distribution which defaults to the minimum
    count in the distribution for unknown values.
    """
    def __init__(self, dist):
        self.update(dist)
        self._total = dist._total
        
        # Prune empty counts
        for key in self.keys():
            if FreqDist.__getitem__(self, key) == 0:
                del self[key]
        
        self._min = min(self.itervalues())
        self._min_prob = self._min / float(self._total)
        self._min_log_prob = log(self._min_prob)
        
    def __getitem__(self, key):
        if key in self:
            return FreqDist.__getitem__(self, key)
        
        return self._min
    
    def prob(self, key):
        if key in self:
            return FreqDist.prob(self, key)
        
        return self._min_prob
    
    def log_prob(self, key):
        if key in self:
            return FreqDist.log_prob(self, key)
        
        return self._min_log_prob
        
    
#----------------------------------------------------------------------------#

class ConditionalFreqDist(dict):
    """
    A model for P(Sample|Condition) for a number of conditions.
    """
    #------------------------------------------------------------------------#
    # PUBLIC METHODS
    #------------------------------------------------------------------------#

    def inc(self, condition, sample, n=1):
        """Increments a count of (sample|condition)."""
        condition_dist = self.get(condition)
        if condition_dist is None:
            condition_dist = self.setdefault(condition, FreqDist())

        condition_dist.inc(sample, n)
        return

    #------------------------------------------------------------------------#

    def prob(self, condition, sample):
        """
        Returns P(sample | condition). An exception is raised for unseen
        conditions.
        """
        condition_dist = self.get(condition)
        if condition_dist is None:
            raise UnknownSymbolError, condition
        else:
            return condition_dist.prob(sample)

    #------------------------------------------------------------------------#

    def log_prob(self, condition, sample):
        """
        Returns log(P(sample | condition)). An exception is raised for
        unseen conditions.
        """
        condition_dist = self.get(condition)
        if condition_dist is None:
            raise UnknownSymbolError, condition
        else:
            return condition_dist.log_prob(sample)

    #------------------------------------------------------------------------#
    
    def candidates(self, condition):
        "Return candidates for the given condition."
        conditionModel = self.get(condition)

        if conditionModel is None:
            # Miss, no such condition recorded.
            return []
        else:
            # Hit, return candidates.
            return conditionModel.candidates()

    #------------------------------------------------------------------------#

    def itercounts(self):
        """
        Returns an interator over all the counts in this model, presented
        as a sequence of (condition, sample, count) tuples.
        """
        for condition, sample_dist in self.iteritems():
            for sample, count in sample_dist.iteritems():
                yield condition, sample, count
        return

    #------------------------------------------------------------------------#

    def invert(self):
        """
        Returns a P(condition|sample) model based off the same counts as
        here.
        """
        newModel = ConditionalFreqDist()
        for condition, sample, count in self.itercounts():
            newModel.inc(sample, condition, count)

        return newModel

    #------------------------------------------------------------------------#

    def dump(self, filename):
        """
        Dump this model to a filename.
        """
        o_stream = sopen(filename, 'w')
        for condition, sample, count in sorted(self.itercounts(),
                key=lambda x: (x[0], -x[2], x[1])):
            condition = _escape_spaces(unicode(condition))
            sample = _escape_spaces(unicode(sample))
            print >> o_stream, u"%s %s %d" % (
                    condition, sample, count
                )
        o_stream.close()
        return

    #------------------------------------------------------------------------#

    def load(self, filename):
        """
        Load counts for this model from a filename.
        """
        i_stream = sopen(filename, 'r')
        for line in i_stream:
            condition, sample, count = line.rstrip().split(_symbol_sep)
            condition = _unescape_spaces(condition)
            sample = _unescape_spaces(sample)
            count = int(count)
            self.inc(condition, sample, count)
        i_stream.close()
        return

    #------------------------------------------------------------------------#

    @staticmethod
    def from_file(filename):
        """
        Alternative constructor. Builds a conditional frequency distribution
        from a file in one line.
        """
        obj = ConditionalFreqDist()
        obj.load(filename)
        return obj

    #------------------------------------------------------------------------#

    def to_condition_dist(self):
        """Generates a frequency distribution of conditions."""
        dist = FreqDist()
        for condition, condition_dist in self.iteritems():
            dist.inc(condition, sum(condition_dist.itervalues()))

        return dist

    #------------------------------------------------------------------------#

    def to_sample_dist(self):
        """Generates a freqency distribution of samples."""
        dist = FreqDist()
        for condition, condition_dist in self.iteritems():
            dist.merge(condition_dist)

        return dist

#----------------------------------------------------------------------------#

def smooth_by_adding_one(freq_dist):
    # XXX this type of smoothing has a particular name (Bell smoothing?) 
    for sample in freq_dist.iterkeys():
        freq_dist[sample] += 1
    return

#----------------------------------------------------------------------------#

class UnknownSymbolError(Exception):
    """
    An error which gets thrown when encountering an unknown character.
    """
    pass

#----------------------------------------------------------------------------#

_symbol_sep = ' '           # The separator we use in our file format
_space_replacement = '_^_'  # The replacement string for dumps

def _escape_spaces(value):
    """
    Esacapes any spaces in the given string with a special value.
    
    >>> _escape_spaces('dog eats cat')
    'dog_^_eats_^_cat'

    >>> _escape_spaces('cow')
    'cow'
    """
    return value.replace(_symbol_sep, _space_replacement)

def _unescape_spaces(value):
    """
    Unescapes any escaped spaces in the string.
    
    >>> _unescape_spaces('dog_^_eats_^_cat')
    'dog eats cat'
    
    >>> _unescape_spaces('cow')
    'cow'
    """
    return value.replace(_space_replacement, _symbol_sep)

def _contains_escape(value):
    """
    Checks the string for the special replacement sequence.
    
    >>> _contains_escape('cow')
    False
    
    >>> _contains_escape('dog_^_eats_^_cat')
    True
    """
    return _space_replacement in value

def sopen(filename, mode='rb', encoding='utf8'):
    """
    Transparently uses compression on the given file based on file
    extension.

    @param filename: The filename to use for the file handle.
    @param mode: The mode to open the file in, e.g. 'r' for read, 'w' for
        write, 'a' for append.
    @param encoding: The encoding to use. Can be set to None to avoid
        using unicode at all.
    """
    read_mode = 'r' in mode
    if read_mode and 'w' in mode:
        raise Exception, "Must be either read mode or write, but not both"

    if filename.endswith('.bz2'):
        stream = bz2.BZ2File(filename, mode)
    elif filename.endswith('.gz'):
        stream = gzip.GzipFile(filename, mode)
    elif filename == '-':
        if read_mode:
            stream = sys.stdin
        else:
            stream = sys.stdout
    else:
        stream = open(filename, mode)

    if encoding not in (None, 'byte'):
        if read_mode:
            return codecs.getreader(encoding)(stream)
        else:
            return codecs.getwriter(encoding)(stream)

    return stream
